/*
 * deviced
 *
 * Copyright (c) 2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <vconf.h>
#include <bundle.h>
#include <eventsystem.h>

#include "core/log.h"
#include "core/list.h"
#include "core/common.h"
#include "core/device-notifier.h"
#include "display/poll.h"
#include "extcon/extcon.h"
#include "apps/apps.h"
#include "usb.h"
#include "usb-tethering.h"

static struct usb_gadget_translator *gadget_translator;
static struct usb_client *usb_client;
static unsigned current_mode;

struct extcon_ops extcon_usb_ops;

static int usb_probe(void)
{
	struct hw_info *info;
	struct hw_common *common;
	int ret;

	/* We need both HALs up and running for operating in USB client mode */

	if (!gadget_translator) {
		ret = hw_get_info(USB_GADGET_DEVICE_ID,
				  (const struct hw_info **)&info);
		if (ret)
			goto no_gadget_hal;

		if (!info->open) {
			_E("Failed to open gadget translator; open(NULL)");
			/* TODO. Shouldn't we free somehow info structure? */
			return -ENODEV;
		}

		ret = info->open(info, NULL, &common);
		if (ret < 0) {
			_E("Failed to open usb gadget translator (%d)", ret);
			return ret;
		}

		gadget_translator = container_of(common,
						 struct usb_gadget_translator,
						 common);
		if (!gadget_translator->id_to_gadget ||
		    !gadget_translator->cleanup_gadget) {
			_E("Invalid gadget translator HAL");
			goto invalid_translator_hal;
		}
	}

	if (!usb_client) {
		ret = hw_get_info(USB_CLIENT_HARDWARE_DEVICE_ID,
				  (const struct hw_info **)&info);
		if (ret)
			goto no_client_hal;

		if (!info->open) {
			_E("Failed to open usb client hw; open(NULL)");
			/* TODO. Shouldn't we free somehow info structure? */
			return -ENODEV;
		}

		ret = info->open(info, NULL, &common);
		if (ret < 0) {
			_E("Failed to open usb client hw (%d)", ret);
			return ret;
		}

		usb_client = container_of(common, struct usb_client, common);

		if (!usb_client->reconfigure_gadget || !usb_client->enable ||
		    !usb_client->disable)
			goto invalid_config_hal;
	}

	return 0;

invalid_config_hal:
	if (usb_client->common.info->close)
		usb_client->common.info->close(&usb_client->common);

no_client_hal:
invalid_translator_hal:
	if (gadget_translator->common.info)
		gadget_translator->common.info->close(&gadget_translator->common);
	gadget_translator = NULL;
no_gadget_hal:
	/* TODO. Maybe some default action here? */
	return -ENOENT;
}

static void usb_release()
{
	if (usb_client)
		if (usb_client->common.info->close)
			usb_client->common.info->close(&usb_client->common);

	if (gadget_translator)
		if (gadget_translator->common.info->close)
			gadget_translator->common.info->close(
				&gadget_translator->common);
}

static int usb_config_init(void)
{
	int ret;
	struct usb_gadget_id gadget_id;
	struct usb_gadget *gadget;

	if (!gadget_translator || !usb_client)
		goto no_hal;

	memset(&gadget_id, 0, sizeof(gadget_id));
	
	gadget_id.function_mask = usb_state_get_selected_mode();
	ret = gadget_translator->id_to_gadget(&gadget_id, &gadget);
	if (ret) {
		_E("Unable to translate id into gadget (%d)", ret);
		return ret;
	}

	usb_client->disable(usb_client);
	ret = usb_client->reconfigure_gadget(usb_client, gadget);
	gadget_translator->cleanup_gadget(gadget);
	if (ret)
		_E("Unable to configure gadget (%d)", ret);

	return ret;
no_hal:
	/* TODO. Maybe some default action here? */
	return -ENODEV;
}

static void usb_config_deinit(void)
{
	if (!usb_client)
		goto no_hal;

	usb_client->disable(usb_client);

	return;
no_hal:
	/* TODO. Maybe some default action here? */
	return;
}

static int usb_config_enable(void)
{
	if (!usb_client)
		goto no_hal;

	return usb_client->enable(usb_client);

no_hal:
	/* TODO. Maybe some default action here? */
	return -ENODEV;
}

static int usb_config_disable(void)
{
	if (!usb_client)
		goto no_hal;

	usb_client->disable(usb_client);

	return 0;
no_hal:
	/* TODO. Maybe some default action here? */
	return -ENODEV;
}

int usb_change_mode(unsigned mode)
{
	struct usb_gadget_id gadget_id;
	struct usb_gadget *gadget;
	int ret;

	if (!gadget_translator || !usb_client)
		goto no_hal;

	memset(&gadget_id, 0, sizeof(gadget_id));
	gadget_id.function_mask = mode;
	
	ret = gadget_translator->id_to_gadget(&gadget_id, &gadget);
	if (ret) {
		_E("Unable to translate id into gadget (%d)", ret);
		return ret;
	}

	usb_client->disable(usb_client);
	ret = usb_client->reconfigure_gadget(usb_client, gadget);
	gadget_translator->cleanup_gadget(gadget);
	if (ret) {
		_E("Unable to configure gadget (%d)", ret);
		goto out;
	}

	current_mode = mode;
out:
	return ret;
no_hal:
	/* TODO. Maybe some default action here? */
	return -ENODEV;
}

static int usb_state_changed(int status)
{
	static int old = -1;	/* to update at the first time */
	unsigned mode;
	int ret;

	_I("USB state is changed from (%d) to (%d)", old, status);

	if (old == status)
		return 0;

	switch (status) {
	case USB_CONNECTED:
		_I("USB cable is connected");
		usb_state_update_state(USB_CONNECTED, USB_FUNCTION_NONE);
		mode = usb_state_get_selected_mode();
		if (mode != current_mode)
			usb_change_mode(mode);

		ret = usb_config_enable();
		if (ret < 0) {
			_E("Failed to enable usb config (%d)", ret);
			break;
		}
		usb_operation_start(mode);
		usb_state_update_state(USB_CONNECTED, mode);
		pm_lock_internal(INTERNAL_LOCK_USB,
				LCD_OFF, STAY_CUR_STATE, 0);
		break;
	case USB_DISCONNECTED:
		_I("USB cable is disconnected");
		usb_operation_stop(usb_state_get_current_mode());
		usb_state_update_state(USB_DISCONNECTED, USB_FUNCTION_NONE);
		
		ret = usb_config_disable();
		if (ret != 0)
			_E("Failed to disable usb config (%d)", ret);
		pm_unlock_internal(INTERNAL_LOCK_USB,
				LCD_OFF, STAY_CUR_STATE);
		break;
	default:
		_E("Invalid USB state(%d)", status);
		return -EINVAL;
	}
	if (ret < 0)
		_E("Failed to operate usb connection(%d)", ret);
	else
		old = status;

	return ret;
}

static void usb_init(void *data)
{
	int ret;

	ret = usb_probe();
	if (ret < 0) {
		_E("USB client cannot be used (%d)", ret);
		return;
	}

	usb_state_retrieve_selected_mode();
	
	ret = usb_config_init();
	if (ret < 0)
		_E("Failed to initialize usb configuation");

	ret = usb_dbus_init();
	if (ret < 0)
		_E("Failed to init dbus (%d)", ret);

	add_usb_tethering_handler();

	ret = usb_state_changed(extcon_usb_ops.status);
	if (ret < 0)
		_E("Failed to update usb status(%d)", ret);
}

static void usb_exit(void *data)
{
	remove_usb_tethering_handler();
	usb_state_update_state(USB_DISCONNECTED, USB_FUNCTION_NONE);
	usb_config_deinit();
	usb_release();
}

struct extcon_ops extcon_usb_ops = {
	.name	= EXTCON_CABLE_USB,
	.init	= usb_init,
	.exit	= usb_exit,
	.update = usb_state_changed,
};

EXTCON_OPS_REGISTER(extcon_usb_ops)
